from typing import Annotated, Any, ClassVar, Callable, Literal, Type, Iterator
from itertools import chain
from functools import cache
from enum import Enum
import inspect

from jinja2 import Environment, PackageLoader, TemplateNotFound
from pydantic import BaseModel, Field, model_validator
from pydantic.fields import FieldInfo
from pydantic import ValidationError

from fastapi.exceptions import RequestValidationError


def is_component(value: Any) -> bool:
    return isinstance(value, BaseComponent)


class BaseComponent(BaseModel):
    # NOTE: If you want to define custom components using your own templates,
    #       override this to your package name (must be stored in `./templates`)
    html_template_package: ClassVar[str] = "quikui"

    @classmethod
    @property
    @cache
    def env(cls):
        env = Environment(
            loader=PackageLoader(cls.html_template_package),
            autoescape=True,
        )
        env.filters.update({"is_component": is_component})
        return env

    @property
    def template(self):
        # NOTE: Will only function with components from this library,
        #       unless `.initialize` is overriden
        if self.env is None:
            raise ValueError(
                "Your component classes need to be initialized with a template"
            )

        template_class = self.__class__
        while issubclass(template_class, BaseComponent):
            try:
                return self.env.get_template(f"{template_class.__name__}.html")

            except TemplateNotFound:
                # NOTE: Potentially dangerous if multi-class heirarchy exists, only uses first
                template_class = template_class.__base__

        raise ValueError(
            f"Component '{self.__class__.__name__}' does not subclass"
            " a component with a valid template."
        )

    def model_dump_html(self, *args, **kwargs) -> str:
        model_dict = self.model_dump(*args, **kwargs)
        return self.template.render(**model_dict)


class Page(BaseComponent):
    title: str
    content: list[BaseComponent]


class Heading(BaseComponent):
    content: str | BaseComponent
    level: Annotated[int, Field(strict=True, ge=1, le=5)] = 1


class Paragraph(BaseComponent):
    content: str | BaseComponent


class Break(BaseComponent):

    def model_dump_html(self) -> str:
        return "<br>"


class TargetType(str, Enum):
    SAME_FRAME = "_self"
    NEW_WINDOW = "_blank"
    PARENT_FRAME = "_parent"
    FULL_BODY = "_top"

    def __str__(self) -> str:
        return self._value_


class Anchor(BaseComponent):
    route: str
    content: str | BaseComponent
    target: TargetType | str = TargetType.SAME_FRAME


class RequestType(str, Enum):
    GET = "get"
    POST = "post"
    PUT = "put"
    PATHC = "patch"
    DELETE = "delete"

    def __str__(self) -> str:
        return self._value_


class Button(BaseComponent):
    route: str
    content: str | BaseComponent
    verb: RequestType = RequestType.GET


class _MultiItemComponent(BaseComponent):
    items: list[str | BaseComponent] = []

    def __init__(self, *args: BaseComponent) -> None:
        if not isinstance(args, (str, BaseComponent)) and (
            isinstance(args, (tuple, list))
            and not all(isinstance(c, (str, BaseComponent)) for c in args)
        ):
            raise ValueError(
                "Must provide either BaseComponent or list of BaseComponents."
            )

        super().__init__(items=list(args))


class Div(_MultiItemComponent):
    pass


class Span(_MultiItemComponent):
    pass


class List(_MultiItemComponent):
    pass


class Image(BaseComponent):
    source: str


class InputType(str, Enum):
    TEXT = "text"
    EMAIL = "email"
    PASSWORD = "password"
    SELECT = "select"
    SUBMIT = "submit"
    RESET = "reset"
    CHECKBOX = "checkbox"
    RADIO = "radio"

    def __str__(self) -> str:
        return self._value_


class FormInput(BaseComponent):
    id: str | None = None
    type: Literal[InputType]
    label: str | BaseComponent | None = None
    add_break: bool = False
    required: bool = True

    @classmethod
    def from_model_field(cls, field_name: str, field_info: FieldInfo) -> "FormInput":
        return cls(
            id=field_name, **field_info.json_schema_extra.get("form_attributes", {})
        )


class TextInput(FormInput):
    type: Literal[InputType] = InputType.TEXT


class EmailInput(FormInput):
    type: Literal[InputType] = InputType.EMAIL


class PasswordInput(FormInput):
    type: Literal[InputType] = InputType.PASSWORD


class InputOption(BaseModel):
    value: str
    label: str | BaseComponent


class InputWithOptions(FormInput):
    options: list[InputOption] = []

    @classmethod
    def from_model_field(cls, field_name: str, field_info: FieldInfo) -> "FormInput":
        form_attributes = field_info.json_schema_extra.get("form_attributes", {})

        # NOTE: Override to allow adding options directly from enum field annotation
        if "options" not in form_attributes and issubclass(field_info.annotation, Enum):
            form_attributes["options"] = [
                InputOption(label=option.value, value=option.value)
                for option in field_info.annotation
            ]

        return cls(id=field_name, **form_attributes)
        # NOTE: Can add options later via `self.options.extend(...)`


class SelectionInput(InputWithOptions):
    type: Literal[InputType] = InputType.SELECT
    # NOTE: Can set these after the initialization
    selected: str | None = None
    disabled: set[str] = {}

    @model_validator(mode="after")
    def check_selected_disabled(self) -> "SelectionInput":
        option_values = set(option.value for option in self.options)

        if self.selected and self.selected not in option_values:
            raise ValidationError("Selected must be one of the current options.")

        if not self.disabled.issubset(option_values):
            raise ValidationError("All disabled options should be current options.")

        return self


class CheckboxInput(FormInput):
    type: Literal[InputType] = InputType.CHECKBOX


class RadioInput(InputWithOptions):
    type: Literal[InputType] = InputType.RADIO


class SubmitForm(FormInput):
    type: Literal[InputType] = InputType.SUBMIT


class ResetForm(FormInput):
    type: Literal[InputType] = InputType.RESET


class Form(BaseComponent):
    id: str | None = None
    verb: RequestType = RequestType.POST
    route: str
    items: list[BaseComponent]


class FormModel(BaseModel):

    @classmethod
    def create_form_input(
        cls,
        field_name: str,
    ) -> FormInput:
        if not (field_info := cls.model_fields.get(field_name)):
            raise ValueError

        if not (input_class := field_info.json_schema_extra.get("form_type")):
            raise ValueError

        return input_class.from_model_field(field_name, field_info)

    @classmethod
    def create_form_items(cls) -> Iterator[BaseComponent]:
        for field_name in cls.model_fields:
            yield cls.create_form_input(field_name)
            yield Break()

        yield SubmitForm()

    @classmethod
    def create_form(
        cls,
        route: str,
        id: str | None = None,
        verb: RequestType = RequestType.POST,
    ) -> Form:
        return Form(
            route=route,
            id=id,
            verb=verb,
            items=list(cls.create_form_items()),
        )


__all__ = [
    BaseComponent.__name__,
    Page.__name__,
    Heading.__name__,
    Paragraph.__name__,
    Break.__name__,
    TargetType.__name__,
    Anchor.__name__,
    RequestType.__name__,
    Button.__name__,
    Div.__name__,
    Span.__name__,
    List.__name__,
    Form.__name__,
    FormInput.__name__,
    InputType.__name__,
    TextInput.__name__,
    EmailInput.__name__,
    PasswordInput.__name__,
    SelectionInput.__name__,
    InputOption.__name__,
    CheckboxInput.__name__,
    RadioInput.__name__,
    SubmitForm.__name__,
    FormModel.__name__,
]



from fastapi import FastAPI, Request, Form, Depends
from enum import Enum
import quikui as qc
from pydantic import BaseModel, Field


app = FastAPI()


@app.get("/")
@app.get("/index.html")
@qc.render_component(is_page=True)
async def index():
    return qc.Page(
        title="Basic Demo App",
        content=[
            qc.Div(
                qc.Heading(content="Basic Component Demo"),
                qc.Paragraph(content="This is a paragraph element."),
                qc.Div(
                    qc.Paragraph(content="This is a paragraph inside a div."),
                    qc.Paragraph(content="This is another paragraph inside a div."),
                    qc.Paragraph(
                        content=qc.Span(
                            "This is a span",
                            qc.Anchor(
                                content="with a link",
                                route="https://google.com",
                                target=qc.TargetType.NEW_WINDOW,
                            ),
                            "to something inside that same div.",
                        )
                    ),
                ),
                qc.Paragraph(
                    content=qc.Span(
                        "This is another span",
                        qc.Anchor(
                            content="with a link",
                            route="/another-page",
                            # NOTE: `.target` defaults to same page
                        ),
                        "to another page, outside the div.",
                    )
                ),
            )
        ],
    )


@app.get("/another-page")
@qc.render_component(is_page=True)
def another_page():
    return qc.Page(
        title="A page with dynamic content",
        content=[
            qc.Heading(content="Using HTMX"),
            qc.Paragraph(
                content=(
                    "This button uses HTMX to dynamically fetch content"
                    " from the server using a GET request."
                )
            ),
            qc.Button(
                route="/dynamic", content="Get Dynamic Content", verb=qc.RequestType.GET
            ),
        ],
    )


class CustomComponent(qc.BaseComponent):
    """To make a renderable component, just subclass this..."""

    text: str = "Some random gibberish!"

    def model_dump_html(self) -> str:
        """...and implement this method"""
        # NOTE: You do not have to subclass a component from the library,
        #       simply use whatever components you'd like to use to render
        return qc.Paragraph(content=self.text).model_dump_html()

    # NOTE: If you override `html_template_package` class variable with your own
    #       package or app, and that contains a `/templates` folder that has a template
    #       with the same name as this class, then it will "auto-render" using that template
    #       without having to override `model_dump_html`


@app.get("/dynamic")
@qc.render_component()
def dynamic_content():
    return CustomComponent()


class CarTypes(Enum):
    NONE_SELECTED = "Select a Car"
    FIAT = "Fiat"
    DODGE = "Dodge"


class CustomForm(qc.FormModel):
    username: str = Field(
        form_type=qc.TextInput,
        form_attributes=dict(label="Username:", add_break=True),
    )
    email: str = Field(
        form_type=qc.EmailInput,
        form_attributes=dict(label="Email:", add_break=True),
        add_break=True,
    )
    password: str = Field(
        form_type=qc.PasswordInput,
        form_attributes=dict(label="Password:", add_break=True),
    )
    save_password: bool = Field(
        form_type=qc.CheckboxInput,
        form_attributes=dict(label="Save Password?"),
    )
    car_choice: CarTypes = Field(
        form_type=qc.RadioInput,
        form_attributes=dict(label="First Choice:"),
    )
    backup_choice: CarTypes = Field(
        form_type=qc.SelectionInput,
        form_attributes=dict(
            label="Second Choice:",
            selected=CarTypes.NONE_SELECTED,
            disabled={CarTypes.NONE_SELECTED},
        ),
    )

    # NOTE: Override `.create_form_items()` with custom form item generator if desired


@app.get("/form")
@qc.render_component(is_page=True)
def form_page():
    return qc.Page(
        title="A page with a form",
        content=[CustomForm.create_form(id="a-form", route="/completed-form")],
    )


@app.post("/completed-form")
@qc.render_component(is_page=True)
async def receive_form(form: CustomForm = Depends(qc.form_handler(CustomForm))):
    return qc.Page(
        title="Form Data",
        content=[
            qc.Heading(content="Form Data"),
            *(
                qc.Paragraph(content=f"{field}: {value}")
                for field, value in form.dict().items()
            ),
        ],
    )



from functools import wraps, partial
import inspect
from typing import Annotated, Any, ClassVar, Callable, Coroutine
from fastapi import (
    HTTPException,
    Depends,
    Header,
    Request,
    status,
    Response,
)
from fastapi.responses import HTMLResponse

from .components import BaseComponent, Page
from .utils import (
    DependsHXRequest,
    append_to_signature,
    execute_maybe_sync_func,
    MaybeAsyncFunc,
    P,
    T,
    get_response,
)


def render_component(
    html_only: bool = False,
    is_page: bool = False,
) -> Callable[[MaybeAsyncFunc[P, T]], Callable[P, Coroutine[None, None, T | Response]]]:

    def decorator(
        func: MaybeAsyncFunc[P, T],
    ) -> Callable[P, Coroutine[None, None, T | Response]]:

        if is_page:

            @wraps(func)
            async def wrapper_render_page(
                *args: P.args,
                __page_request: Request,
                **kwargs: P.kwargs,
            ) -> T | Response:
                result = await execute_maybe_sync_func(func, *args, **kwargs)
                if isinstance(result, Response):
                    return result

                elif __page_request is None:  # or not isinstance(result, Page):
                    raise HTTPException(
                        status.HTTP_422_UNPROCESSABLE_ENTITY,
                        "This component must be rendered from a page request.",
                    )

                response = get_response(kwargs)
                return HTMLResponse(
                    result.model_dump_html(),
                    headers=None if response is None else response.headers,
                )

            return append_to_signature(
                wrapper_render_page,
                inspect.Parameter(
                    "__page_request",
                    inspect.Parameter.KEYWORD_ONLY,
                    annotation=Request,
                ),
            )

        else:

            @wraps(func)
            async def wrapper_render_if_htmx(
                *args: P.args,
                __hx_request: DependsHXRequest,
                **kwargs: P.kwargs,
            ) -> T | Response:
                if html_only and __hx_request is None:
                    raise HTTPException(
                        status.HTTP_400_BAD_REQUEST,
                        "This route can only process HTMX requests.",
                    )

                result = await execute_maybe_sync_func(func, *args, **kwargs)
                if __hx_request is None or isinstance(result, Response):
                    return result

                elif isinstance(result, BaseComponent):
                    result = result.model_dump_html()

                elif not isinstance(result, str):
                    # NOTE: Should not happen if library is used properly
                    raise HTTPException(
                        status.HTTP_422_UNPROCESSABLE_ENTITY,
                        "Function does not render properly.",
                    )

                response = get_response(kwargs)
                return HTMLResponse(
                    result,
                    headers=None if response is None else response.headers,
                )

            return append_to_signature(
                wrapper_render_if_htmx,
                inspect.Parameter(
                    "__hx_request",
                    inspect.Parameter.KEYWORD_ONLY,
                    annotation=DependsHXRequest,
                ),
            )

    return decorator



import inspect
from asyncio import iscoroutinefunction
from collections.abc import Callable, Mapping
from typing import Any, cast, Annotated, Coroutine
from typing import ParamSpec, Protocol, TypeVar

from fastapi import Request, Response


from fastapi import Response, Request, Header, Depends
from fastapi.concurrency import run_in_threadpool

P = ParamSpec("P")
T = TypeVar("T")
MaybeAsyncFunc = Callable[P, T] | Callable[P, Coroutine[Any, Any, T]]


def get_hx_request(
    request: Request, hx_request: Annotated[str | None, Header()] = None
) -> Request | None:
    """
    FastAPI dependency that returns the current request if it is an HTMX one,
    i.e. it contains an `"HX-Request: true"` header.
    """
    return request if hx_request == "true" else None


DependsHXRequest = Annotated[Request | None, Depends(get_hx_request)]


def append_to_signature(func: Callable, *params: inspect.Parameter) -> Callable:
    """
    Appends the given parameters to the *end* of the signature of the given function.

    Notes:
        - This method does not change the function's arguments, it only makes FastAPI's
        dependency resolution system recognize inserted parameters.
        - This is *not* a general purpose method, it is strongly recommended to only
        append keyword-only parameters that have "unique" names that are unlikely to
        be already in the function's signature.

    Arguments:
        func: The function whose signature should be extended.
        params: The parameters to add to the function's signature.

    Returns:
        The received function with an extended `__signature__`.
    """
    signature = inspect.signature(func)
    func.__signature__ = signature.replace(parameters=(*signature.parameters.values(), *params))  # type: ignore[attr-defined]
    return func


async def execute_maybe_sync_func(
    func: MaybeAsyncFunc[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    """
    Executes the given function in a thread if it's a sync one, or in the current asyncio
    event loop if it's an async one.

    Arguments:
        func: The function to execute.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.
    """
    if iscoroutinefunction(func):
        return await func(*args, **kwargs)  # type: ignore[no-any-return]

    return await run_in_threadpool(cast(Callable[P, T], func), *args, **kwargs)


def get_response(kwargs: Mapping[str, Any]) -> Response | None:
    """
    Returns the first `Response` instance from the values in `kwargs` (if there is one).

    Arguments:
        kwargs: The keyword arguments from which the `Response` should be returned.
    """
    for val in kwargs.values():
        if isinstance(val, Response):
            return val

    return None
